html {
  width: 100%;
  height: 100%;
  background: #FAFAFA;
}

body {
  height: 100%;
  width: 100%;
  margin: 0;
  box-sizing: border-box;
  font-family: Helvetica, Arial, sans-serif;
  padding: 0px 10px;
  padding-top: 3rem;
  padding-bottom: 1rem;
}

.navbar {
  position: fixed;
  top: 0;
  left: 0;
  right: 0;
  z-index: 1;
}

.navbar ul {
  list-style-type: none;
  margin: 0;
  padding: 0;
  overflow: hidden;
  background-color: #EEE;
}

.navbar li {
  float: left;
  font-size: 17px;
  transition: .3s background-color;
}

.navbar li.active {
  background-color: rgba(234, 170, 109, 0.7);
}

.navbar li a {
  display: block;
  color: black;
  padding: 11px 16px;
  text-decoration: none;
}

.navbar li:hover {
  background-color: #eaaa6d;
}

#dask-logo img {
  height: 28px;
  padding: 5px 15px;
}

#dask-logo a {
  padding: 0px;
}

#navbar-toggle-icon {
  float: right;
}

#navbar-toggle-icon a {
  display: none;
}

#navbar-toggle-icon img {
  height: 22px;
}

@media screen and (max-width: 650px) {
  .navbar li:not(#dask-logo):not(#navbar-toggle-icon) a {
    display: none;
  }
  #navbar-toggle-icon a {
    display: block;
  }
}

@media screen and (max-width: 650px) {
  .navbar.responsive li:not(#navbar-toggle-icon) {
    float: none;
  }
  .navbar.responsive li:not(#dask-logo):not(#navbar-toggle-icon) a {
    display: block;
    text-align: left;
  }
  .navbar.responsive #navbar-toggle-icon a {
    position: absolute;
    right: 0;
    top: 0;
  }
}

.bk-root .bk-toolbar-box .bk-toolbar-right {
  top: 4px;
  right: 4px;
}

.content {
  width: 100%;
  height: 100%;
}
